import pickle
from logging import getLogger

from qtpy import QtCore, QtWidgets

import bw2data as bd

from activity_browser import signals, settings
from activity_browser.ui import icons

from activity_browser.ui.menu_bar import MenuBar

log = getLogger(__name__)


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self):
        super().__init__()

        self.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        self.setWindowTitle("Activity Browser")
        self.setWindowIcon(icons.qicons.ab)
        self.setDockNestingEnabled(True)

        # Layout: extra items outside main layout
        self.menu_bar = MenuBar(self)
        self.setMenuBar(self.menu_bar)

        self.connect_signals()

    def closeEvent(self, event):
        """
        Save the state of the main window when it is closed.
        """
        # Save the state of the main window
        self.writeState(bd.projects.dir)
        super().closeEvent(event)

    def sync(self):
        from activity_browser.layouts import panes

        self.clearPanes()

        data = self.getState(bd.projects.dir)

        unique_panes = [pane for pane in panes.registered_panes if pane.unique]

        for pane_state in data.get("panes", []):
            self.restorePane(pane_state)
            if pane_state.get("class") in unique_panes:
                unique_panes.remove(pane_state.get("class"))

        for pane in unique_panes:
            self.addPane(pane(parent=self))

        success = self.restoreState(data.get("state"), 0)
        log.debug(f"Restored main window state: {success}")

        self.setWindowTitle(f"Activity Browser - {bd.projects.current}")


    def connect_signals(self):
        # Keyboard shortcuts
        signals.project.changed.connect(self.on_project_changed)

    def clearPanes(self):
        for pane in self.panes():
            pane.deleteLater()

    def panes(self):
        """
        Return a list of all panes in the main window.
        """
        from activity_browser.ui import widgets
        return self.findChildren(widgets.ABAbstractPane)

    def on_project_changed(self, new, old):
        self.writeState(old.dir)
        self.sync()

    def restorePane(self, pane_state: dict):
        """
        Restore a pane in the main window based on its saved state.

        Args:
            pane_state (dict): A dictionary containing the class and state of the pane to restore.
                - "class": The class of the pane to be restored.
                - "state": The saved state of the pane.

        Returns:
            None
        """
        pane_class = pane_state.get("class")
        if pane_class is None:
            return

        log.debug(f"Restoring pane {pane_class.__name__}")

        try:
            # Attempt to create an instance of the pane using its saved state
            pane_instance = pane_class.fromState(pane_state.get("state"), self)
        except Exception as e:
            # Log an error if the pane cannot be restored
            log.error(f"Error restoring pane {pane_class.__name__}: {e}")
            return

        self.addPane(pane_instance)

    def addPane(self, pane: QtWidgets.QWidget):
        """
        Add a pane to the main window.

        Args:
            pane (QtWidgets.QWidget): The pane to add.

        Returns:
            None
        """
        dockwidget = pane.getDockWidget(self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, dockwidget)
        self.menu_bar.view_menu.addAction(dockwidget.toggleViewAction())

        pane.sync()

    def writeState(self, directory):
        pane_data = []
        for pane in self.panes():
            pane_data.append({
                "class": pane.__class__,
                "state": pane.saveState(),
            })
        own_data = self.saveState(0)

        data = {
            "state": own_data,
            "panes": pane_data,
        }

        # Save the state of the main window
        path = directory.joinpath("activity_browser\\main_window_state.pickle")
        with open(path, "wb") as f:
            pickle.dump(data, f)

    def getState(self, directory) -> dict:
        """
        Get the state of the main window.
        """
        path = directory.joinpath("activity_browser\\main_window_state.pickle")
        if path.exists():
            with open(path, "rb") as f:
                data = pickle.load(f)
        else:
            data = {
                "state": self.saveState(0),
            }
        return data


    def set_titlebar(self):
        self.setWindowTitle(f"Activity Browser - {bd.projects.current}")

    def dialog_on_exception(self, exception: Exception):
        QtWidgets.QMessageBox.critical(
            self,
            f"An error occurred: {type(exception).__name__}",
            f"An error occurred, check the logs for more information \n\n {str(exception)}",
            QtWidgets.QMessageBox.Ok,
        )



try:
    from bw2data.backends.peewee.database import *
except ModuleNotFoundError:
    # we're running bw25
    from bw2data.backends.base import *

from activity_browser.mod.patching import patch_superclass, patched
from activity_browser.signals import (qactivity_list, qdatabase_list,
                                      qexchange_list)

from .proxies import Activity, ActivityDataset, Exchange, ExchangeDataset


@patch_superclass
class SQLiteBackend(SQLiteBackend):

    @property
    def changed(self):
        """
        Shorthand for connecting to the database QUpdater. Developers can instantiate a Database from bw2data and
        connect directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qdatabase_list.get_or_create(self).changed

    @property
    def deleted(self):
        """
        Shorthand for connecting to the database QUpdater. Developers can instantiate a Database from bw2data and
        connect directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qdatabase_list.get_or_create(self).deleted

    def delete(self, *args, **kwargs) -> None:
        # get all affected activities and exchanges that have QUpdater counterparts (i.e. have signals attached to them)
        acts = [
            (Activity(ActivityDataset.get_by_id(qact["id"])), qact)
            for qact in qactivity_list
            if qact["database"] == self.name
        ]

        excs = [
            (Exchange(ExchangeDataset.get_by_id(qexc["id"])), qexc)
            for qexc in qexchange_list
            if qexc["input_database"] == self.name
            or qexc["output_database"] == self.name
        ]

        # execute the patched function for standard functionality
        patched[SQLiteBackend]["delete"](self, *args, **kwargs)

        # emit the deleted db, affected activities, and affected exchanges
        [
            qdb.emitLater("changed", self)
            for qdb in qdatabase_list
            if qdb["name"] == self.name
        ]
        [
            qdb.emitLater("deleted", self)
            for qdb in qdatabase_list
            if qdb["name"] == self.name
        ]

        for act, qact in acts:
            qact.emitLater("changed", act)
        for act, qact in acts:
            qact.emitLater("deleted", act)

        for exc, qexc in excs:
            qexc.emitLater("changed", exc)
        for exc, qexc in excs:
            qexc.emitLater("deleted", exc)

    def get(self, code) -> Activity:
        return patched[SQLiteBackend]["get"](self, code)



from bw2data.method import *

from activity_browser.signals import qmethod_list

from ..patching import patch_superclass, patched


@patch_superclass
class Method(Method):
    @property
    def changed(self):
        """
        Shorthand for connecting to the method QUpdater. Developers can instantiate a Method from bw2data and connect
        directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qmethod_list.get_or_create(self).changed

    @property
    def deleted(self):
        """
        Shorthand for connecting to the method QUpdater. Developers can instantiate a Method from bw2data and connect
        directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qmethod_list.get_or_create(self).deleted

    def write(self, data, process=True):
        # execute the patched function for standard functionality
        patched[Method]["write"](self, data, process)

        # emit for any corresponding qmethod that exists in qmethod_list (each method that has widgets connected to it)
        [
            qmthd.emitLater("changed", self)
            for qmthd in qmethod_list
            if qmthd["name"] == self.name
        ]

    def deregister(self):
        # execute the patched function for standard functionality
        patched[Method]["deregister"](self)

        # emit for any corresponding qmethod that exists in qmethod_list (each method that has widgets connected to it)
        [
            qmthd.emitLater("deleted", self)
            for qmthd in qmethod_list
            if qmthd["name"] == self.name
        ]
        [
            qmthd.emitLater("changed", self)
            for qmthd in qmethod_list
            if qmthd["name"] == self.name
        ]

    # extending Brightway Functionality
    def load_dict(self) -> dict:
        """
        Extending Brightway functionality by enabling users to get a method dictionary with elementary flow keys as keys
        instead of a list. This makes it easier to find characterization factors.
        """
        return {cf[0]: cf[1] for cf in self.load()}

    def write_dict(self, data: dict, process=True):
        """
        Extending Brightway functionality by enabling users to write a method dictionary with elementary flow keys as
        keys instead of a list. This makes it easier to edit characterization factors.
        """
        self.write(list(data.items()), process)



from bw2analyzer import ContributionAnalysis

from typing import Optional
import numpy as np

class ABContributionAnalysis(ContributionAnalysis):
    """Activity Browser version of bw2analyzer.ContributionAnalysis"""
    def sort_array(self, data: np.array, limit: float = 25, limit_type: str = "number", total: Optional[float] = None) -> np.array:
        """Activity Browser version of bw2analyzer.ContributionAnalysis.sort_array.

        Should be removed once https://github.com/brightway-lca/brightway2-analyzer/pull/32 is merged.
        See PR above on why we overwrite this function.
        """
        if not total:
            abs_total_flag = True
            total = np.abs(data).sum()
        else:
            abs_total_flag = False

        if total == 0 and limit_type == "cum_percent":
            raise ValueError("Cumulative percentage cannot be calculated to a total of 0, use a different limit type or total")

        if limit_type not in ("number", "percent", "cum_percent"):
            raise ValueError(f"limit_type must be either 'number', 'percent' or 'cum_percent' not '{limit_type}'.")
        if limit_type  in ("percent", "cum_percent"):
            if not 0 < limit <= 1:
                raise ValueError("Percentage limits > 0 and <= 1.")

        results = np.hstack(
            (data.reshape((-1, 1)), np.arange(data.shape[0]).reshape((-1, 1)))
        )

        if limit_type == "number":
            # sort and cut off at limit
            return results[np.argsort(np.abs(data))[::-1]][:limit, :]
        elif limit_type == "percent":
            # identify good values, drop rest and sort
            limit = (np.abs(data) >= (abs(total) * limit))
            results = results[limit, :]
            return results[np.argsort(np.abs(results[:, 0]))[::-1]]
        elif limit_type == "cum_percent" and abs_total_flag:
            # if we would apply this on the 'correct' order, this would stop just before the limit,
            # we want to be on or the first step over the limit.
            results = results[np.argsort(np.abs(data))]  # sort low to high impact
            cumsum = np.cumsum(np.abs(results[:, 0])) / abs(total)
            limit = (cumsum >= (1 - limit))  # find items under limit
            return results[limit, :][::-1]  # drop items under limit and set correct order
        elif limit_type == "cum_percent" and not abs_total_flag:
            # iterate over positive and negative values until limit is achieved or surpassed.
            results = results[np.argsort(np.abs(data))][::-1]
            pos_neg = [  # split into positive and negative sections
                results[results[:, 0] > 0],
                results[results[:, 0] < 0],
            ]
            # iterate over positive and negative sections
            for i, arr in enumerate(pos_neg):
                c = 0
                # iterate over array until we have equalled or surpassed limit
                for j, row in enumerate(arr):
                    c += abs(row[0] / total)
                    if c >= limit:
                        break
                arr = arr[:min(j + 1, len(arr)), :]
                pos_neg[i] = arr

            results = np.concatenate(pos_neg)  # rebuild into 1 array
            return results[np.argsort(np.abs(results[:, 0]))][::-1]  # sort values



try:
    from bw2data.backends.peewee.proxies import *
except ModuleNotFoundError:
    # we're running bw25
    from bw2data.backends.proxies import *

from bw2data import Database, get_activity

from activity_browser.mod.patching import patch_superclass, patched
from activity_browser.signals import (qactivity_list, qdatabase_list,
                                      qexchange_list)


@patch_superclass
class Exchanges(Exchanges):

    def delete(self, allow_in_sourced_project: bool = False):
        # find only the exchanges that have qexchange counterparts within ourselves
        exc_query = ExchangeDataset.id << [qexc["id"] for qexc in qexchange_list]
        exc_args = self._args + [exc_query]
        excs = [Exchange(doc) for doc in ExchangeDataset.select().where(*exc_args)]

        # find only the input or output activities that have qactivity counterparts

        # get all qactivity keys
        act_keys = [(qact["database"], qact["code"]) for qact in qactivity_list]

        # gather affected output activities

        # construct a preliminary query using only the output code (this should get us very far)
        act_query = ExchangeDataset.output_code << [key[1] for key in act_keys]

        # combine with the existing query
        act_args = self._args + [act_query]

        # execute query to a set, and only select activities that are in act_keys
        acts = {
            get_activity((doc.output_database, doc.output_code))
            for doc in ExchangeDataset.select().where(*act_args)
            if (doc.output_database, doc.output_code) in act_keys
        }

        # gather affected input activities

        # same process as above but for input_code and input_database
        act_query = ExchangeDataset.input_code << [key[1] for key in act_keys]
        act_args = self._args + [act_query]
        acts.update(
            {
                get_activity((doc.input_database, doc.input_code))
                for doc in ExchangeDataset.select().where(*act_args)
                if (doc.input_database, doc.input_code) in act_keys
            }
        )

        # use the activities set to create a database set as well
        dbs = set([act["database"] for act in acts])

        # execute the patched function for standard functionality
        patched[Exchanges]["delete"](self, allow_in_sourced_project)

        # emitting change through any existing exchange QUpdaters
        for exc in excs:
            [
                qexc.emitLater("changed", exc)
                for qexc in qexchange_list
                if qexc["id"] == exc._document.id
            ]
            [
                qexc.emitLater("deleted", exc)
                for qexc in qexchange_list
                if qexc["id"] == exc._document.id
            ]

        # emitting change through any existing activity QUpdaters
        for act in acts:
            [
                qact.emitLater("changed", act)
                for qact in qactivity_list
                if qact["id"] == act._document.id
            ]

        # emitting change through any existing database QUpdaters
        for db_name in dbs:
            [
                qdb.emitLater("changed", Database(db_name))
                for qdb in qdatabase_list
                if qdb["name"] == db_name
            ]


@patch_superclass
class Activity(Activity):

    @property
    def changed(self):
        """
        Shorthand for connecting to the activity QUpdater. Developers can instantiate an Activity from bw2data and
        connect directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qactivity_list.get_or_create(self).changed

    @property
    def deleted(self):
        """
        Shorthand for connecting to the activity QUpdater. Developers can instantiate an Activity from bw2data and
        connect directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qactivity_list.get_or_create(self).deleted

    def save(self, signal: bool = True, data_already_set: bool = False,
             force_insert: bool = False) -> None:
        from activity_browser.bwutils.metadata import AB_metadata

        # execute the patched function for standard functionality
        patched[Activity]["save"](self, signal, data_already_set, force_insert)

        # this is called already within the patched function, but needs to be recalled now the data is actually updated
        databases.set_modified(self["database"])

        # should eventually be replaced
        AB_metadata.update_metadata(self.key)

        # exchanges cannot be changed through the activity proxy save function

        # emitting change through any existing qactivities (should be 1 or None)
        [
            qact.emitLater("changed", self)
            for qact in qactivity_list
            if qact["id"] == self._document.id
        ]

        # emitting change through an existing qdatabases (should be 1 or None)
        db = Database(self["database"])
        [
            qdb.emitLater("changed", db)
            for qdb in qdatabase_list
            if qdb["name"] == self["database"]
        ]

    def delete(self, signal: bool = True):
        from activity_browser.bwutils.metadata import AB_metadata

        # execute the patched function for standard functionality
        patched[Activity]["delete"](self, signal)

        databases.set_modified(self["database"])

        # this is leading to a lot of calls and should eventually be dealt with
        AB_metadata.update_metadata(self.key)

        # exchange deletions will emit for themselves

        # emitting change through any existing qactivities (should be 1 or None)
        [
            qact.emitLater("changed", self)
            for qact in qactivity_list
            if qact["id"] == self._document.id
        ]
        [
            qact.emitLater("deleted", self)
            for qact in qactivity_list
            if qact["id"] == self._document.id
        ]

        # emitting change through an existing qdatabases (should be 1 or None)
        db = Database(self["database"])
        [
            qdb.emitLater("changed", db)
            for qdb in qdatabase_list
            if qdb["name"] == self["database"]
        ]


@patch_superclass
class Exchange(Exchange):

    def __init__(self, document=None, **kwargs):
        # execute the patched function for standard functionality
        patched[Exchange]["__init__"](self, document, **kwargs)

        # we need to keep track of changed input and output activities, as they will have to be signalled too when the
        # exchange is saved.
        self.moved_IO = []

    @property
    def changed(self):
        """
        Shorthand for connecting to the activity QUpdater. Developers can instantiate an Activity from bw2data and
        connect directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qexchange_list.get_or_create(self).changed

    @property
    def deleted(self):
        """
        Shorthand for connecting to the activity QUpdater. Developers can instantiate an Activity from bw2data and
        connect directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qexchange_list.get_or_create(self).deleted

    def save(self, signal: bool = True, data_already_set: bool = False,
             force_insert: bool = False):
        # execute the patched function for standard functionality
        patched[Exchange]["save"](self, signal, data_already_set, force_insert)

        # emitting change through any existing qexchanges (should be 1 or None)
        [
            qexc.emitLater("changed", self)
            for qexc in qexchange_list
            if qexc["id"] == self._document.id
        ]

        # collecting unique activities and databases that have changed
        acts = set()
        dbs = set()

        acts.add(self.input)
        acts.add(self.output)
        acts.update(self.moved_IO)

        # emitting change through any existing qactivities
        for activity in acts:
            dbs.add(activity["database"])
            [
                qact.emitLater("changed", activity)
                for qact in qactivity_list
                if qact["id"] == activity._document.id
            ]

        # emitting change through any existing qdatabases
        for db_name in dbs:
            db = Database(db_name)
            [
                qdb.emitLater("changed", db)
                for qdb in qdatabase_list
                if qdb["name"] == db_name
            ]

        self.moved_IO.clear()

    def delete(self, signal: bool = True):
        # execute the patched function for standard functionality
        patched[Exchange]["delete"](self, signal)

        # emitting change and deletion through any existing qexchanges (should be 1 or None)
        [
            qexc.emitLater("changed", self)
            for qexc in qexchange_list
            if qexc["id"] == self._document.id
        ]
        [
            qexc.emitLater("deleted", self)
            for qexc in qexchange_list
            if qexc["id"] == self._document.id
        ]

        # emitting change for any existing qactivities (should be 1 or None)
        [
            qact.emitLater("changed", self.input)
            for qact in qactivity_list
            if qact["id"] == self.input._document.id
        ]
        [
            qact.emitLater("changed", self.output)
            for qact in qactivity_list
            if qact["id"] == self.output._document.id
        ]

        # emitting change for related databases
        in_db = Database(self.input["database"])
        [
            qdb.emitLater("changed", in_db)
            for qdb in qdatabase_list
            if qdb["name"] == in_db.name
        ]

        out_db = Database(self.output["database"])
        [
            qdb.emitLater("changed", out_db)
            for qdb in qdatabase_list
            if qdb["name"] == out_db.name
        ]



import os

from bw2data.backends import *

from .base import SQLiteBackend
from .proxies import Activity, ActivityDataset, Exchange, ExchangeDataset

try:
    from bw2data.backends.peewee import sqlite3_lci_db
except ModuleNotFoundError:
    # we're running bw25
    os.environ["AB_BW25"] = "True"
    pass

Node = Activity
Edge = Exchange



from pathlib import Path
import os

from bw2data.project import *

try:
    from bw2data.backends.peewee import SubstitutableDatabase
except ModuleNotFoundError:
    # we're running Brightway 25 so import accordingly
    from bw2data.backends import SubstitutableDatabase

from activity_browser.signals import qprojects

from ..patching import patch_superclass, patched


@patch_superclass
class ProjectManager(ProjectManager):
    @property
    def current_changed(self):
        """
        Shorthand for connecting to the method QUpdater. Developers can import projects from bw2data and connect
        directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qprojects.current_changed

    @property
    def list_changed(self):
        """
        Shorthand for connecting to the method QUpdater. Developers can import projects from bw2data and connect
        directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qprojects.list_changed

    def set_current(self, name, writable=True, update=True):
        # execute the patched function for standard functionality
        patched[ProjectManager]["set_current"](self, name, writable, update)

        # emit that the current project has changed through the qUpdater
        qprojects.emitLater("current_changed")

    def create_project(self, name=None, **kwargs):
        # execute the patched function for standard functionality
        patched[ProjectManager]["create_project"](self, name, **kwargs)

        # emit that the project list has changed through the qUpdater
        qprojects.emitLater("list_changed")

    def copy_project(self, new_name, switch=True):
        # execute the patched function for standard functionality
        patched[ProjectManager]["copy_project"](self, new_name, switch)

        # emit that the project list has changed through the qUpdater
        qprojects.emitLater("list_changed")

    def delete_project(self, name=None, delete_dir=False):
        # execute the patched function for standard functionality
        patched[ProjectManager]["delete_project"](self, name, delete_dir)

        # emit that the project list has changed through the qUpdater
        qprojects.emitLater("list_changed")

    # extending functionality
    @property
    def base_dir(self) -> str:
        """
        We keep using the protected property anyway, so why not make it available.
        """
        return projects._base_data_dir

    def switch_dir(self, path: Path) -> None:
        """
        Switch the brightway2 project directory to the given path.
        """
        # change the paths to the given dir
        if not isinstance(path, Path):
            path = Path(path)

        projects._base_data_dir = path
        projects._base_logs_dir = path / "logs"
        projects._base_logs_dir.mkdir(exist_ok=True, parents=True)

        # open (or otherwise create) the projects.db at the given location
        projects.db = SubstitutableDatabase(
            projects._base_data_dir / "projects.db", [ProjectDataset]
        )

        # emit that the current project has changed through the qUpdater
        qprojects.emitLater("current_changed")


projects: ProjectManager = projects



from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod.bw2io import remote
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class ProjectNewRemote(ABAction):
    """
    ABAction to create a new project from a remote template.
    """

    icon = qicons.add
    text = "New project from remote"
    tool_tip = "Make a new project from remote template"

    @staticmethod
    @exception_dialogs
    def run(project_key: str):
        name, ok = QtWidgets.QInputDialog.getText(
            application.main_window,
            "Create project from remote",
            "Name of new project:" + " " * 25,
        )

        if not ok or not name:
            return

        if name in bd.projects:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible.",
                "A project with this name already exists.",
            )
            return

        remote.install_project(project_key, name)
        bd.projects.set_current(name)



from typing import Any, List

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class ExchangeFormulaRemove(ABAction):
    """
    ABAction to clear the formula's of one or more exchanges.
    """

    icon = qicons.delete
    text = "Clear formula(s)"

    @staticmethod
    @exception_dialogs
    def run(exchanges: List[Any]):
        for exchange in exchanges:
            try:
                del exchange["formula"]
                exchange.save()
            except KeyError:
                # formula not in the exchange
                continue



from argparse import Namespace

from bw2data.parameters import (
    ActivityParameter,
    DatabaseParameter,
    Group,
    GroupDependency,
    ParameterBase,
    ParameterizedExchange,
    ParameterManager,
    ProjectParameter,
    parameters,
)

from activity_browser.signals import qparameter_list, qparameters

from ..patching import patch_attribute, patch_superclass, patched


class PatchedParameterBase(ParameterBase):
    """
    ParameterBase is already patched by Peewee, so we need to work more precisely by patching specific attributes
    instead of the entire class at once.
    """

    @patch_attribute(ParameterBase, "changed")
    @property
    def changed(self):
        """
        Shorthand for connecting to the parameter QUpdater. Developers can instantiate a parameter from bw2data and
        connect directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qparameter_list.get_or_create(self).changed

    @patch_attribute(ParameterBase, "deleted")
    @property
    def deleted(self):
        """
        Shorthand for connecting to the parameter QUpdater. Developers can instantiate a parameter from bw2data and
        connect directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qparameter_list.get_or_create(self).deleted

    @patch_attribute(ParameterBase, "key")
    @property
    def key(self):
        """
        Extension to get a unique key for each parameter based on group name and own name. Useful for finding the right
        QUpdater
        """
        if isinstance(self, ProjectParameter):
            return "project", self.name
        elif isinstance(self, DatabaseParameter):
            return self.database, self.name
        elif isinstance(self, ActivityParameter):
            return self.group, self.name
        return None

    @patch_attribute(ParameterBase, "delete")
    @classmethod
    def delete(cls):
        """
        We're actually patching peewee here, because the information we need is contained in the .where() call that
        follows the delete call. So we return a Namespace in which .where() is the function described below. When .where
        is called we extract what we need, emit the correct signals, and return the patched .delete().where() call.
        """

        def where(*args):
            """Patched .where() function. We use the *args to get the parameters that will be deleted from the
            database using a select() call"""
            # call the database with the where *args supplied by the user
            for param in cls.select().where(*args):
                # emit that any connected params will be changed and deleted
                [
                    qprm.emitLater("changed", param)
                    for qprm in qparameter_list
                    if qprm["key"] == param.key
                ]
                [
                    qprm.emitLater("deleted", param)
                    for qprm in qparameter_list
                    if qprm["key"] == param.key
                ]

            # also emit the overall qparameters
            qparameters.emitLater("parameters_changed")

            # return by calling the patched function to restore normal functionality
            return patched[ParameterBase]["delete"].__func__(cls).where(*args)

        def execute():
            """Patched .execute() function. Meaning all params are deleted. Signal accordingly"""
            # collect al params from the database
            for param in cls.select():
                # emit that any connected params will be changed and deleted
                [
                    qprm.emitLater("changed", param)
                    for qprm in qparameter_list
                    if qprm["key"] == param.key
                ]
                [
                    qprm.emitLater("deleted", param)
                    for qprm in qparameter_list
                    if qprm["key"] == param.key
                ]

            # also emit the overall qparameters
            qparameters.emitLater("parameters_changed")

            # return by calling the patched function to restore normal functionality
            return patched[ParameterBase]["delete"].__func__(cls).execute()

        return Namespace(where=where, execute=execute)

    @patch_attribute(ParameterBase, "update")
    @classmethod
    def update(cls, *update_args, **update_kwargs):
        """
        We're actually patching peewee here, because the information we need is contained in the .where() call that
        follows the update call. So we return a Namespace in which .where() is the function described below. When .where
        is called we extract what we need, emit the correct signals, and return the patched .update().where() call.
        """

        def where(*where_args, **where_kwargs):
            """Patched .where() function. We use the *args to get the parameters that will be updated from the
            database using a select() call"""
            # call the database with the where *args supplied by the user
            for param in cls.select().where(*where_args):
                # emit that any connected params will be changed
                [
                    qprm.emitLater("changed", param)
                    for qprm in qparameter_list
                    if qprm["key"] == param.key
                ]
                [
                    qprm.emitLater("deleted", param)
                    for qprm in qparameter_list
                    if qprm["key"] == param.key
                ]

            # also emit the overall qparameters
            qparameters.emitLater("parameters_changed")

            # return by calling the patched function to restore normal functionality
            return (
                patched[ParameterBase]["update"]
                .__func__(cls, *update_args, **update_kwargs)
                .where(*where_args, **where_kwargs)
            )

        def execute():
            for param in cls.select():
                # emit that any connected params will be changed
                [
                    qprm.emitLater("changed", param)
                    for qprm in qparameter_list
                    if qprm["key"] == param.key
                ]
                [
                    qprm.emitLater("deleted", param)
                    for qprm in qparameter_list
                    if qprm["key"] == param.key
                ]

            # also emit the overall qparameters
            qparameters.emitLater("parameters_changed")

            # return by calling the patched function to restore normal functionality
            return (
                patched[ParameterBase]["update"]
                .__func__(cls, *update_args, **update_kwargs)
                .execute()
            )

        return Namespace(where=where, execute=execute)

    @patch_attribute(ParameterBase, "insert_many")
    @classmethod
    def insert_many(cls, *args, **kwargs):
        # emit to overall qparamaters
        qparameters.emitLater("parameters_changed")

        # return by calling the patched function to restore normal functionality
        return patched[ParameterBase]["insert_many"].__func__(cls, *args, **kwargs)

    @patch_attribute(ParameterBase, "save")
    def save(self, **kwargs):
        # call the patched function to have normal functionality
        patched[ParameterBase]["save"](self, **kwargs)

        # signal the changed parameter if it has signals connected to it
        [
            qprm.emitLater("changed", self)
            for qprm in qparameter_list
            if qprm["key"] == self.key
        ]

        # always signal through the qparameters if a parameter has changed
        qparameters.emitLater("parameters_changed")


@patch_superclass
class ParameterManager(ParameterManager):
    @property
    def parameters_changed(self):
        """
        Shorthand for connecting to the parameters QUpdater. Developers can connect directly to the parameters
        singleton, instead of importing the related QUpdater via activity_browser.signals
        """
        return qparameters.parameters_changed


parameters: ParameterManager = parameters



from bw2data.meta import *

from activity_browser.signals import qcalculation_setups, qdatabases, qmethods

from ..patching import patch_superclass, patched


@patch_superclass
class Databases(Databases):
    """
    Apparently the databases metaclass is kinda broken. As in, it doesn't update when the underlying databases change in
    e.g. number of activities. Also, setting a database as modified is done before the database is actually modified,
    meaning that callbacks relying on the modified database will fail. This should be fixed within Brightway...
    """

    @property
    def metadata_changed(self):
        """
        Shorthand for connecting to the qdatabases QUpdater. Developers can import 'databases' from bw2data and connect
        directly, instead of importing the related QUpdater via activity_browser.signals
        """
        return qdatabases.metadata_changed

    def flush(self, signal: bool = True):
        """
        Emit that the databases have changed when it's data is flushed to disk a.k.a. saved.
        """
        # execute the patched function for standard functionality
        patched[Databases]["flush"](self, signal)
        # emit that the databases metadata have changed through the qUpdater
        qdatabases.emitLater("metadata_changed")


@patch_superclass
class CalculationSetups(CalculationSetups):
    @property
    def metadata_changed(self):
        """
        Shorthand for connecting to the qcalculation-setups QObject. Developers can import 'databases' from bw2data
        and connect directly, instead of importing the related QObject via activity_browser.signals
        """
        return qcalculation_setups.metadata_changed

    def flush(self, signal: bool = True):
        """
        Emit that the calculation setups have changed when it's data is flushed to disk a.k.a. saved.
        """
        # execute the patched function for standard functionality
        patched[CalculationSetups]["flush"](self, signal)
        # emit that the calculation setups metadata have changed through the qUpdater
        qcalculation_setups.emitLater("metadata_changed")


@patch_superclass
class Methods(Methods):
    @property
    def metadata_changed(self):
        """
        Shorthand for connecting to the qmethods QObject. Developers can import 'databases' from bw2data and connect
        directly, instead of importing the related QObject via activity_browser.signals
        """
        return qmethods.metadata_changed

    def flush(self, signal: bool = True):
        """
        Emit that the methods have changed when it's data is flushed to disk a.k.a. saved.
        """
        # execute the patched function for standard functionality
        patched[Methods]["flush"](self, signal)
        # emit that the methods metadata have changed through the qUpdater
        qmethods.emitLater("metadata_changed")


# reimport the patched singletons, but adding the type-hint so IDE's know what functionality has been added/patched
databases: Databases = databases
calculation_setups: CalculationSetups = calculation_setups
methods: Methods = methods



import bw2data.errors as errors
import bw2data.configuration as configuration
from bw2data import *

from .backends import Edge, Node
from .meta import calculation_setups, databases, methods
from .method import Method
from .parameters import parameters
from .project import projects
from .utils import get_activity



from bw2data.utils import *

from bw2data.backends import Activity


def get_activity(key) -> Activity:
    """Re-init of get_activity to show the IDE that we're returning a patched activity"""
    import bw2data

    return bw2data.utils.get_activity(key)



# -*- coding: utf-8 -*-
"""Classes related to the cutoff options menu in contributions tabs.

These classes contain all menu items required to modify the cutoffs of the MLCA results. The
CutoffMenu class is responsible for assembling the menu. Each different menu item is contained in
its separate class.
"""

from collections import namedtuple
from typing import Union

import numpy as np
from qtpy import QtCore
from qtpy.QtCore import QLocale, Qt, Signal, Slot
from qtpy.QtGui import QDoubleValidator, QIntValidator
from qtpy.QtWidgets import (QButtonGroup, QHBoxLayout, QLabel, QLineEdit,
                               QPushButton, QRadioButton, QSlider, QVBoxLayout,
                               QWidget)

from activity_browser.ui import widgets


# These tuples are used in referring to the two Types and three Labels used
Types = namedtuple("types", ("relative", "topx"))
Labels = namedtuple("labels", ("unit", "min", "max"))


class CutoffMenu(QWidget):
    """This class assembles the cutoff menu from the other classes in this module."""

    slider_change = Signal()

    def __init__(self, parent=None, cutoff_value=0.01, limit_type="percent"):
        super().__init__(parent)
        self.cutoff_value = cutoff_value
        self.limit_type = limit_type

        locale = QLocale(QLocale.English, QLocale.UnitedStates)
        locale.setNumberOptions(QLocale.RejectGroupSeparator)
        self.validators = Types(
            QDoubleValidator(0.001, 100.0, 1, self), QIntValidator(0, 50, self)
        )
        self.validators.relative.setLocale(locale)
        self.validators.topx.setLocale(locale)
        self.buttons = Types(QRadioButton("Relative"), QRadioButton("Top #"))
        self.buttons.relative.setChecked(True)
        self.buttons.relative.setToolTip(
            "This cut-off type shows the selected top percentage of contributions (for example the \
top 10% contributors)"
        )
        self.buttons.topx.setToolTip(
            "This cut-off type shows the selected top number of contributions (for example the top \
5 contributors)"
        )
        self.button_group = QButtonGroup()
        self.button_group.addButton(self.buttons.relative, 0)
        self.button_group.addButton(self.buttons.topx, 1)
        self.sliders = Types(LogarithmicSlider(self), QSlider(Qt.Horizontal, self))
        self.sliders.relative.setToolTip(
            "This slider sets the selected percentage of contributions\
 to be shown"
        )
        self.sliders.topx.setToolTip(
            "This slider sets the selected number of contributions to be \
shown"
        )
        self.units = Types("% of total", "top #")
        self.labels = Labels(QLabel(), QLabel(), QLabel())
        self.cutoff_slider_line = QLineEdit()
        self.cutoff_slider_line.setToolTip(
            "This box can set a precise cut-off value for the \
contributions to be shown"
        )
        self.cutoff_slider_line.setLocale(locale)
        self.cutoff_slider_lft_btn = QPushButton("<")
        self.cutoff_slider_lft_btn.setToolTip(
            "This button moves the cut-off value one increment"
        )
        self.cutoff_slider_rght_btn = QPushButton(">")
        self.cutoff_slider_rght_btn.setToolTip(
            "This button moves the cut-off value one increment"
        )

        self.debounce_slider = QtCore.QTimer()
        self.debounce_slider.setInterval(750)
        self.debounce_slider.setSingleShot(True)

        self.debounce_text = QtCore.QTimer()
        self.debounce_text.setInterval(300)
        self.debounce_text.setSingleShot(True)

        self.make_layout()
        self.connect_signals()

    def connect_signals(self):
        """Connect the signals of the menu."""
        # Cut-off types
        self.buttons.relative.toggled.connect(self.cutoff_type_check)
        self.cutoff_slider_lft_btn.clicked.connect(self.cutoff_increment_left_check)
        self.cutoff_slider_rght_btn.clicked.connect(self.cutoff_increment_right_check)

        self.debounce_slider.timeout.connect(self.initiate_slider_change)
        self.debounce_text.timeout.connect(self.initiate_text_change)

        self.sliders.relative.valueChanged.connect(self.debounce_slider.start)
        self.sliders.topx.valueChanged.connect(self.debounce_slider.start)
        self.cutoff_slider_line.textChanged.connect(self.debounce_text.start)

    def initiate_slider_change(self):
        if self.is_relative:
            self.cutoff_slider_relative_check("sl")
        else:
            self.cutoff_slider_topx_check("sl")

    def initiate_text_change(self):
        if self.is_relative:
            self.cutoff_slider_relative_check("le")
        else:
            self.cutoff_slider_topx_check("le")

    @property
    def is_relative(self) -> bool:
        """Check if relative button is checked."""
        return self.buttons.relative.isChecked()

    @Slot(name="incrementLeftCheck")
    def cutoff_increment_left_check(self):
        """Move the slider 1 increment to left when left button is clicked."""
        if self.is_relative:
            num = int(self.sliders.relative.value())
            self.sliders.relative.setValue(num + 1)
        else:
            num = int(self.sliders.topx.value())
            self.sliders.topx.setValue(num - 1)

    @Slot(name="incrementRightCheck")
    def cutoff_increment_right_check(self):
        """Move the slider 1 increment to right when right button is clicked."""
        if self.is_relative:
            num = int(self.sliders.relative.value())
            self.sliders.relative.setValue(num - 1)
        else:
            num = int(self.sliders.topx.value())
            self.sliders.topx.setValue(num + 1)

    @Slot(bool, name="isRelativeToggled")
    def cutoff_type_check(self, toggled: bool) -> None:
        """Dependent on cutoff-type, set the right labels.

        Slot connected to the relative radiobutton, the state of that button determines:
        - which sliders are visible
        - the unit shown
        - minimum and maximum
        - limit_type
        """
        self.sliders.topx.setVisible(not toggled)
        self.sliders.relative.setVisible(toggled)

        self.sliders.relative.blockSignals(True)
        self.sliders.topx.blockSignals(True)
        self.cutoff_slider_line.blockSignals(True)
        if toggled:
            self.labels.unit.setText(self.units.relative)
            self.labels.min.setText("100%")
            self.labels.max.setText("0.001%")
            self.limit_type = "percent"
            self.cutoff_slider_line.setValidator(self.validators.relative)
        else:
            self.labels.unit.setText(self.units.topx)
            self.labels.min.setText(str(self.sliders.topx.minimum()))
            self.labels.max.setText(str(self.sliders.topx.maximum()))
            self.limit_type = "number"
            self.cutoff_slider_line.setValidator(self.validators.topx)
        self.sliders.relative.blockSignals(False)
        self.sliders.topx.blockSignals(False)
        self.cutoff_slider_line.blockSignals(False)

    @Slot(str, name="sliderRelativeCheck")
    def cutoff_slider_relative_check(self, editor: str):
        """If 'relative' selected, change the plots and tables to reflect the slider/line-edit."""
        if not self.is_relative:
            return
        cutoff = 0.01

        # If called by slider
        if editor == "sl":
            self.cutoff_slider_line.blockSignals(True)
            cutoff = abs(self.sliders.relative.log_value)
            self.cutoff_slider_line.setText(str(cutoff))
            self.cutoff_slider_line.blockSignals(False)

        # if called by line edit
        elif editor == "le":
            self.sliders.relative.blockSignals(True)
            if self.cutoff_slider_line.text() == "-":
                cutoff = 0.001
                self.cutoff_slider_line.setText("0.001")
            elif self.cutoff_slider_line.text() == "":
                cutoff = 0.001
            else:
                cutoff = abs(float(self.cutoff_slider_line.text()))

            if cutoff > 100:
                cutoff = 100
                self.cutoff_slider_line.setText(str(cutoff))
            self.sliders.relative.log_value = float(cutoff)
            self.sliders.relative.blockSignals(False)

        self.cutoff_value = cutoff / 100
        self.slider_change.emit()

    @Slot(str, name="sliderTopXCheck")
    def cutoff_slider_topx_check(self, editor: str):
        """If 'top #' selected, change the plots and tables to reflect the slider/line-edit."""
        if self.is_relative:
            return
        cutoff = 2

        # If called by slider
        if editor == "sl":
            self.cutoff_slider_line.blockSignals(True)
            cutoff = abs(int(self.sliders.topx.value()))
            self.cutoff_slider_line.setText(str(cutoff))
            self.cutoff_slider_line.blockSignals(False)

        # if called by line edit
        elif editor == "le":
            self.sliders.topx.blockSignals(True)
            if self.cutoff_slider_line.text() == "-":
                cutoff = self.sliders.topx.minimum()
                self.cutoff_slider_line.setText(str(self.sliders.topx.minimum()))
            elif self.cutoff_slider_line.text() == "":
                cutoff = self.sliders.topx.minimum()
            else:
                cutoff = abs(int(self.cutoff_slider_line.text()))

            if cutoff > self.sliders.topx.maximum():
                cutoff = self.sliders.topx.maximum()
                self.cutoff_slider_line.setText(str(cutoff))
            self.sliders.topx.setValue(int(cutoff))
            self.sliders.topx.blockSignals(False)

        self.cutoff_value = int(cutoff)
        self.slider_change.emit()

    def make_layout(self):
        """Assemble the layout of the cutoff menu.

        Construct the layout for the cutoff menu widget. The initial layout is set to 'relative'.
        """
        layout = QHBoxLayout()

        # Cut-off types
        cutoff_type = QVBoxLayout()
        cutoff_type_label = QLabel("Cut-off type")

        # Cut-off slider
        cutoff_slider = QVBoxLayout()
        cutoff_slider_set = QVBoxLayout()
        cutoff_slider_label = QLabel("Cut-off level")
        self.sliders.relative.setInvertedAppearance(True)
        self.sliders.topx.setMinimum(1)
        self.sliders.topx.setMaximum(50)
        self.sliders.topx.setValue(self.cutoff_value)
        self.sliders.relative.log_value = self.cutoff_value
        cutoff_slider_minmax = QHBoxLayout()
        self.labels.min.setText("100%")
        self.labels.max.setText("0.001%")
        self.labels.unit.setText("%  of total")
        cutoff_slider_ledit = QHBoxLayout()
        self.cutoff_slider_line.setValidator(self.validators.relative)
        self.cutoff_slider_lft_btn.setMaximumWidth(15)
        self.cutoff_slider_rght_btn.setMaximumWidth(15)

        # Assemble types
        cutoff_type.addWidget(cutoff_type_label)
        cutoff_type.addWidget(self.buttons.relative)
        cutoff_type.addWidget(self.buttons.topx)

        # Assemble slider set
        self.sliders.topx.setVisible(False)
        cutoff_slider_set.addWidget(cutoff_slider_label)
        cutoff_slider_minmax.addWidget(self.labels.min)
        cutoff_slider_minmax.addWidget(self.sliders.relative)
        cutoff_slider_minmax.addWidget(self.sliders.topx)
        cutoff_slider_minmax.addWidget(self.labels.max)
        cutoff_slider_set.addLayout(cutoff_slider_minmax)

        cutoff_slider_ledit.addWidget(self.cutoff_slider_line)
        cutoff_slider_ledit.addWidget(self.cutoff_slider_lft_btn)
        cutoff_slider_ledit.addWidget(self.cutoff_slider_rght_btn)
        cutoff_slider_ledit.addWidget(self.labels.unit)
        cutoff_slider_ledit.addStretch(1)

        cutoff_slider.addLayout(cutoff_slider_set)
        cutoff_slider.addLayout(cutoff_slider_ledit)

        # Assemble cut-off menu
        layout.addLayout(cutoff_type)
        layout.addWidget(widgets.ABVLine(self))
        layout.addLayout(cutoff_slider)
        layout.addStretch()

        self.setLayout(layout)


class LogarithmicSlider(QSlider):
    """Makes a QSlider object that behaves logarithmically.

    Inherits from QSlider. This class uses the property `log_value` getter and setter to modify
    the QSlider through the `value` and `setValue` methods.
    """

    def __init__(self, parent=None):
        super().__init__(Qt.Horizontal, parent)
        self.setMinimum(1)
        self.setMaximum(100)

    @property
    def log_value(self) -> Union[int, float]:
        """Read (slider) and modify from 1-100 to 0.001-100 logarithmically with relevant rounding.

        This function converts the 1-100 values and modifies these to 0.001-100 on a logarithmic
        scale. Rounding is done based on magnitude.
        """

        # Logarithmic math refresher:
        # BOP = Base, Outcome Power;
        # log(B)(O) = P --> log(2)(64) = 6  ||  log(10)(1000) = 3
        #       B^P = O -->        2^6 = 64 ||           10^3 = 1000

        value = float(self.value())
        log_val = np.log10(value)
        power = log_val * 2.5 - 3
        ret_val = np.power(10, power)
        test_val = np.log10(ret_val)

        if test_val < -1:
            return ret_val.round(3)
        elif test_val < -0:
            return ret_val.round(2)
        elif test_val < 1:
            return ret_val.round(1)
        else:
            return ret_val.astype(int)

    @log_value.setter
    def log_value(self, value: float) -> None:
        """Modify value from 0.001-100 to 1-100 logarithmically and set slider to value."""
        value = int(float(value) * np.power(10, 3))
        log_val = np.log10(value).round(3)
        set_val = log_val * 20
        self.setValue(set_val)
